/*
 *  EKED-M1 : Editor for Korg M1 synth; prog_ed.c : program editor
 *  Copyright (C) 1995 Steven M. Eker (Steven.Eker@brunel.ac.uk)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stddef.h>
#include <osbind.h>		/* needed for Vsync() */
#include <string.h>
#include <gemfast.h>
#include <aesbind.h>
#include <vdibind.h>
#include "gm/gem_man.h"
#include "eked-m1.h"
#include "defs.h"
#include "types.h"
#include "externs.h"

#define T_WID		50
#define WIDTH2TIME(w)	(((w) == 0) ? 0 : 2 * (w) - 1)
#define TIME2WIDTH(t)	(((t) + 1) / 2)
#define HEIGHT2SNUM(h)	(((h) == 0) ? 0 : 2 * (h) + (((h) > 0) ? 1 : -1))
#define SNUM2HEIGHT(n)	(((n) >= 0) ? (n) / 2 : -((-(n)) / 2))
#define HEIGHT2UNUM(h)	(h)
#define UNUM2HEIGHT(n)	(n)
#define NO_GRAPH	(-2)	/* -1 used by objc_find()! */

/*
 *	Graph control point data structure
 */
typedef struct {
  int c_x;		/* x offset from GEM graph object */
  int c_y;		/* y offset from GEM graph object */
#define FIXED		(-1)
#define FOLLOW		(-2)
  int x_ob;		/* GEM object for x parameter or FIXED or FOLLOW */
  int y_ob;		/* GEM object for y parameter or FIXED or FOLLOW */
} CTRL_POINT;

/*
 *	Graph data structure
 */
typedef struct {
  int x_axis_x1;	/* x axis */
  int x_axis_x2;
  int x_axis_y;
  int y_axis_x;		/* y axis 1 (note on) */
  int y_axis_y1;
  int y_axis_y2;	/* y axis 2 (note off) */
  int y_axis2_x;
  int y_min;		/* minimum y value for control point */
  int y_max;		/* maximum y value for control point */
  int n_points;		/* number of control points */
  CTRL_POINT point[7];	/* control point array */
  int vsync_flag;	/* flag to force Vsync() call before redraw */
} GRAPH;

static BITBLK *wave_image[4];

static OBJECT *pitch_mod;
static OBJECT *pitch1;
static OBJECT *pitch2;
static OBJECT *filter_mod;
static OBJECT *filter1;
static OBJECT *filter2;
static OBJECT *misc;
static OBJECT *amp1;
static OBJECT *amp2;
static OBJECT *effect_route;
static OBJECT *effect1;
static OBJECT *effect2;

static BANK *edit_bank;			/* bank of program being edited */
static int number;			/* number of program being edited */
static int in_use;			/* editor in use flag */
static int graph_handle;		/* vdi handle for drawing graphs */
static int work_w;			/* width of window contents */
static int work_h;			/* height of window contents */
static BYTE data[PROGRAM_SIZE];		/* private copy of program */
static BYTE compare[PROGRAM_SIZE];	/* another copy for compare op */

static void prog_setup();
static void display(int x_pos, int y_pos, GRECT *clip, long usr_val);
static int action(int handle, E_TYPE type, void *event,
                  int x_pos, int y_pos, long usr_val);
static void save_prog(BYTE *p);
static void do_click(int handle, int m_x, int m_y);
static void setup(int x_pos, int y_pos);
static void gpara_setup(OBJECT *tree, int graph_ob,
                        PARAMETER table[], int size);
static void gpara_edit(int handle, OBJECT *tree, int graph_ob,
                       PARAMETER table[], int size, int m_x, int m_y);
static void graph_edit(int handle, OBJECT *tree, int ob,
                       PARAMETER table[], int size, int m_x, int m_y);
static void graph_drag(int handle, OBJECT *tree, int ob,
                       int x_offset, int y_offset,
                       CTRL_POINT coord[], int n_move, int link,
                       VRECT *bound, GRECT *area);
static int graph_draw(PARMBLK *p);
static PARAMETER *find_parameter(PARAMETER table[], int size, int ob);
static void para_setup(OBJECT *tree, PARAMETER table[], int size);
static int misc_buttons(int handle, int ob);

void pe_init()
{
  static GRAPH pitch1_graph = {
    4, 187, 56,		8, 4, 108, 128,		7, 105,		 6,
    {
      {8,		66,	FIXED,		P1_ST_LEV},
      {8+T_WID,	10,	P1_AT_TIME,	P1_AT_LEV},
      {8+2*T_WID,	56,	P1_DC_TIME,	FIXED},
      {128,	56,	FIXED,		FIXED},
      {128+T_WID,	20,	P1_RL_TIME,	P1_RL_LEV},
      {187,	20,	FIXED,		FOLLOW}
    }
  };
  
  static GRAPH pitch2_graph = {
    4, 187, 56,		8, 4, 108, 128,		7, 105,		 6,
    {
      {8,		105,	FIXED,		P2_ST_LEV},
      {8+T_WID,	7,	P2_AT_TIME,	P2_AT_LEV},
      {8+2*T_WID,	56,	P2_DC_TIME,	FIXED},
      {128,	56,	FIXED,		FIXED},
      {128+T_WID,	80,	P2_RL_TIME,	P2_RL_LEV},
      {187,	80,	FIXED,		FOLLOW}
    }
  };
  
  static GRAPH filter1_graph = {
    4, 235, 56,		 8, 4, 108, 176,	7, 105,		 7,
    {
      {8,		56,	FIXED,		FIXED},
      {8+T_WID,	90,	F1_AT_TIME,	F1_AT_LEV},
      {8+2*T_WID,	56,	F1_DC_TIME,	F1_DC_LEV},
      {8+3*T_WID,	45,	F1_SS_TIME,	F1_SS_LEV},
      {176,	45,	FIXED,		FOLLOW},
      {176+T_WID,	40,	F1_RL_TIME,	F1_RL_LEV},
      {235,	40,	FIXED,		FOLLOW}
    }
  };
  
  static GRAPH filter2_graph = {
    4, 235, 56,		 8, 4, 108, 176,	7, 105,		 7,
    {
      {8,		56,	FIXED,		FIXED},
      {8+T_WID,	7,      F2_AT_TIME,	F2_AT_LEV},
      {8+2*T_WID,	105,	F2_DC_TIME,	F2_DC_LEV},
      {8+3*T_WID,	20,	F2_SS_TIME,	F2_SS_LEV},
      {176,	20,	FIXED,		FOLLOW},
      {176+T_WID,	40,	F2_RL_TIME,	F2_RL_LEV},
      {235,	40,	FIXED,		FOLLOW}
    }
  };
  
  static GRAPH amp1_graph = {
    4, 235, 104,		 8, 4, 108, 176,	5, 104,		 6,
    {
      {8,		104,	FIXED,		FIXED},
      {8+T_WID,	5,	A1_AT_TIME,	A1_AT_LEV},
      {8+2*T_WID,	104,	A1_DC_TIME,	A1_DC_LEV},
      {8+3*T_WID,	80,	A1_SS_TIME,	A1_SS_LEV},
      {176,	80,	FIXED,		FOLLOW},
      {176+T_WID,	104, 	A1_RL_TIME,	FIXED}
    }
  };
  
  static GRAPH amp2_graph = {
    4, 235, 104,		 8, 4, 108, 176,	5, 104,		 6,
    {
      {8,		104,	FIXED,		FIXED},
      {8+T_WID,	5,	A2_AT_TIME,	A2_AT_LEV},
      {8+2*T_WID,	104,	A2_DC_TIME,	A2_DC_LEV},
      {8+3*T_WID,	80,	A2_SS_TIME,	A2_SS_LEV},
      {176,	80,	FIXED,		FOLLOW},
      {176+T_WID,	104, 	A2_RL_TIME,	FIXED}
    }
  };

  static USERBLK pitch1_ub = {(void *) &graph_draw, (long) &pitch1_graph};
  static USERBLK pitch2_ub = {(void *) &graph_draw, (long) &pitch2_graph};
  static USERBLK filter1_ub = {(void *) &graph_draw, (long) &filter1_graph};
  static USERBLK filter2_ub = {(void *) &graph_draw, (long) &filter2_graph};
  static USERBLK amp1_ub = {(void *) &graph_draw, (long) &amp1_graph};
  static USERBLK amp2_ub = {(void *) &graph_draw, (long) &amp2_graph};

  graph_handle = gm_opnvwk();
  vsf_color(graph_handle, WHITE);
  vsm_type(graph_handle, 4);
  vsm_color(graph_handle, BLACK);
  vsl_udsty(graph_handle, 0xAAAA);

  rm_gaddr(R_TREE, PITCH_MOD, &pitch_mod);
  rm_gaddr(R_TREE, PITCH1, &pitch1);
  rm_gaddr(R_TREE, PITCH2, &pitch2);
  rm_gaddr(R_TREE, FILTER_MOD, &filter_mod);
  rm_gaddr(R_TREE, FILTER1, &filter1);
  rm_gaddr(R_TREE, FILTER2, &filter2);
  rm_gaddr(R_TREE, MISC, &misc);
  rm_gaddr(R_TREE, AMP1, &amp1);
  rm_gaddr(R_TREE, AMP2, &amp2);
  rm_gaddr(R_TREE, EFFECT_ROUTE, &effect_route);
  rm_gaddr(R_TREE, EFFECT1, &effect1);
  rm_gaddr(R_TREE, EFFECT2, &effect2);

  if(char_h >= 16){
    wave_image[0] = (BITBLK *) filter_mod[FM_WAVE].ob_spec;
    wave_image[1] = (BITBLK *) pitch_mod[PM_WAVE].ob_spec;
    rm_gaddr(R_IMAGEDATA, DOWNSAW, wave_image + 2);
    rm_gaddr(R_IMAGEDATA, SQUARE, wave_image + 3);
  }
  else{
    rm_gaddr(R_IMAGEDATA, TRIANGLE_L, wave_image);
    rm_gaddr(R_IMAGEDATA, UPSAW_L, wave_image + 1);
    rm_gaddr(R_IMAGEDATA, DOWNSAW_L, wave_image + 2);
    rm_gaddr(R_IMAGEDATA, SQUARE_L, wave_image + 3);
  }

  pitch1[P1_GRAPH].ob_spec = (long) &pitch1_ub;
  pitch2[P2_GRAPH].ob_spec = (long) &pitch2_ub;
  filter1[F1_GRAPH].ob_spec = (long) &filter1_ub;
  filter2[F2_GRAPH].ob_spec = (long) &filter2_ub;
  amp1[A1_GRAPH].ob_spec = (long) &amp1_ub;
  amp2[A2_GRAPH].ob_spec = (long) &amp2_ub;

  work_w = pitch_mod[0].ob_width + filter_mod[0].ob_width + misc[0].ob_width +
           effect_route[0].ob_width;
  work_h = pitch_mod[0].ob_height + pitch1[0].ob_height + pitch2[0].ob_height;
}

void pe_exit()
{
  v_clsvwk(graph_handle);
}

void pe_window(BANK *bank, int nr)
{
  static char title[] = "Edit Bank 0 Program 00";
  int t;

  if(in_use){
    fm_alert(1, PROG_ED_ALERT);
    return;
  }
  edit_bank = bank;
  number = nr;
  memcpy(data, PROG_DATA(bank, nr), (size_t) PROGRAM_SIZE);
  prog_setup();
  misc[MS_COMPARE].ob_state &= ~SELECTED;
  misc[MS_RESET].ob_state &= ~DISABLED;
  misc[MS_SAVE].ob_state &= ~DISABLED;
  title[10] = '0' + bank->bank_nr;
  title[sizeof(title) - 3] = '0' + nr / 10;
  title[sizeof(title) - 2] = '0' + nr % 10;
  t = wm_open(NAME | CLOSER | FULLER | MOVER | SIZER |
              UPARROW | DNARROW | VSLIDE |
              LFARROW | RTARROW | HSLIDE,
              (GRECT *) 0, title, work_w, work_h, char_w, char_h,
              &display, &action, 0L);
  if(t < 0){
    fm_alert(1, WINDOW_ALERT);
    return;
  }
  in_use = TRUE;
}

static void prog_setup()
{
  para_setup(pitch_mod, pm_table, PM_SIZE);
  para_setup(filter_mod, fm_table, FM_SIZE);
  para_setup(misc, ms_table, MS_SIZE);
  gpara_setup(pitch1, P1_GRAPH, p1_table, P1_SIZE);
  gpara_setup(amp1, A1_GRAPH, a1_table, A1_SIZE);
  gpara_setup(filter1, F1_GRAPH, f1_table, F1_SIZE);
  gpara_setup(pitch2, P2_GRAPH, p2_table, P2_SIZE);
  gpara_setup(amp2, A2_GRAPH, a2_table, A2_SIZE);
  gpara_setup(filter2, F2_GRAPH, f2_table, F2_SIZE);
  eroute_setup(effect_route, data + 38);
  epara_setup(effect1, 0, data + 38);
  epara_setup(effect2, 1, data + 38);
}

static void display(int x_pos, int y_pos, GRECT *clip, long usr_val)
{
  setup(x_pos, y_pos);
  objc_draw(pitch_mod, 0, MAX_DEPTH, clip->g_x, clip->g_y,
            clip->g_w, clip->g_h);
  objc_draw(pitch1, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(pitch2, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(filter_mod, 0, MAX_DEPTH, clip->g_x, clip->g_y,
            clip->g_w, clip->g_h);
  objc_draw(filter1, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(filter2, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(misc, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(amp1, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(amp2, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(effect_route, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w,
            clip->g_h);
  objc_draw(effect1, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
  objc_draw(effect2, 0, MAX_DEPTH, clip->g_x, clip->g_y, clip->g_w, clip->g_h);
}

static int action(int handle, E_TYPE type, void *event,
                  int x_pos, int y_pos, long usr_val)
{
  BYTE *p;

  switch(type){
  case E_CLEANUP:
    m1_ctrl(handle, EXITING, 0, 0, (BYTE *) NULL);
    in_use = FALSE;
    break;
  case E_MESSAGE:
    if(((int *) event)[0] == WM_CLOSED){
      p = (misc[MS_COMPARE].ob_state & SELECTED) ? compare : data;
      if(memcmp(p, PROG_DATA(edit_bank, number), (size_t) PROGRAM_SIZE) != 0){ 
        switch(fm_alert(1, PROG_QUIT_ALERT)){
        case 1:	/* save */
          save_prog(p);
          break;
        case 3:	/* cancel */
          return R_NOACT;
        }
      }
    }
    break;
  case E_CLICK:
    m1_ctrl(handle, 0, EDIT_PROG_MODE, number, data);
    setup(x_pos, y_pos);
    do_click(handle, ((CLICK *) event)->m_x, ((CLICK *) event)->m_y);
    break;
  }
  return R_NORMAL;
}

static void save_prog(BYTE *p)
{
  static int msg[] = {UPDATE_ITEM, 0};

  memcpy(PROG_DATA(edit_bank, number), p, (size_t) PROGRAM_SIZE);
  if(edit_bank->prog_win != 0){
    msg[1] = number;
    wm_send_user(edit_bank->prog_win, (long) &msg);
  }
}

static void do_click(int handle, int m_x, int m_y)
{
  if(m_x < misc[0].ob_x){
    if(m_x < filter_mod[0].ob_x){
      if(m_y < pitch1[0].ob_y)
        gpara_edit(handle, pitch_mod, NO_GRAPH, pm_table, PM_SIZE, m_x, m_y);
      else if(m_y < pitch2[0].ob_y)
        gpara_edit(handle, pitch1, P1_GRAPH, p1_table, P1_SIZE, m_x, m_y);
      else if(data[10] == 1)	/* only recognise in double osc mode */
        gpara_edit(handle, pitch2, P2_GRAPH, p2_table, P2_SIZE, m_x, m_y);
    }
    else{
      if(m_y < pitch1[0].ob_y)
        gpara_edit(handle, filter_mod, NO_GRAPH, fm_table, FM_SIZE, m_x, m_y);
      else if(m_y < pitch2[0].ob_y)
        gpara_edit(handle, filter1, F1_GRAPH, f1_table, F1_SIZE, m_x, m_y);
      else if(data[10] == 1)	/* only recognise in double osc mode */
        gpara_edit(handle, filter2, F2_GRAPH, f2_table, F2_SIZE, m_x, m_y);
    }
  }
  else{
    if(m_x < effect_route[0].ob_x){
      if(m_y < pitch1[0].ob_y)
        gpara_edit(handle, misc, NO_GRAPH, ms_table, MS_SIZE, m_x, m_y);
      else if(m_y < pitch2[0].ob_y)
        gpara_edit(handle, amp1, A1_GRAPH, a1_table, A1_SIZE, m_x, m_y);
      else if(data[10] == 1)	/* only recognise in double osc mode */
        gpara_edit(handle, amp2, A2_GRAPH, a2_table, A2_SIZE, m_x, m_y);
    }
    else{
      if(m_y < effect1[0].ob_y)
        eroute_edit(handle, effect_route, (data[10] == 1) ? 23 : 14, m_x, m_y,
                    data + 38);
      else if(m_y < effect2[0].ob_y){
        effect_edit(handle, effect1, 0, (data[10] == 1) ? 23 : 14, m_x, m_y,
                    data + 38);
      }
      else{
        effect_edit(handle, effect2, 1, (data[10] == 1) ? 23 : 14, m_x, m_y,
                    data + 38);
      }
    }
  }
}

static void setup(int x_pos, int y_pos)
{
  int x1 = x_pos + pitch_mod[0].ob_width;
  int x2 = x1 + filter_mod[0].ob_width;
  int x3 = x2 + misc[0].ob_width;
  int y1 = y_pos + pitch_mod[0].ob_height;
  int y2 = y1 + pitch1[0].ob_height;

  pitch_mod[0].ob_x = x_pos;
  pitch_mod[0].ob_y = y_pos;
  pitch1[0].ob_x = x_pos;
  pitch1[0].ob_y = y1;
  pitch2[0].ob_x = x_pos;
  pitch2[0].ob_y = y2;
  filter_mod[0].ob_x = x1;
  filter_mod[0].ob_y = y_pos;
  filter1[0].ob_x = x1;
  filter1[0].ob_y = y1;
  filter2[0].ob_x = x1;
  filter2[0].ob_y = y2;
  misc[0].ob_x = x2;
  misc[0].ob_y = y_pos;
  amp1[0].ob_x = x2;
  amp1[0].ob_y = y1;
  amp2[0].ob_x = x2;
  amp2[0].ob_y = y2;
  effect_route[0].ob_x = x3;
  effect_route[0].ob_y = y_pos;
  effect1[0].ob_x = x3;
  effect1[0].ob_y = y1;
  effect2[0].ob_x = x3;
  effect2[0].ob_y = y2;
}

static void gpara_setup(OBJECT *tree, int graph_ob,
                        PARAMETER table[], int size)
{
  GRAPH *g = (GRAPH *) (((USERBLK *) tree[graph_ob].ob_spec)->ub_parm);
  CTRL_POINT *point = g->point;
  int i, t;
  PARAMETER *p;

  for(i = 0; i < g->n_points; i++){
    if(point[i].x_ob > FIXED){
      p = find_parameter(table, size, point[i].x_ob);
      t = data[p->offset];
      point[i].c_x = point[i - 1].c_x + TIME2WIDTH(t);
    }
    if(point[i].y_ob > FIXED){
      p = find_parameter(table, size, point[i].y_ob);
      t = data[p->offset];
      t = g->x_axis_y - ((p->type == UNUM) ? UNUM2HEIGHT(t) : SNUM2HEIGHT(t));
      point[i].c_y = t;
      if(i + 1 < g->n_points && point[i + 1].y_ob == FOLLOW)
        point[i + 1].c_y = t;
    }
  }
  para_setup(tree, table, size);
}

static void gpara_edit(int handle, OBJECT *tree, int graph_ob,
                       PARAMETER table[], int size, int m_x, int m_y)
{
  int ob = objc_find(tree, 0, MAX_DEPTH, m_x, m_y);
  GRAPH *g;
  CTRL_POINT *point;
  int i, j, t, update;
  PARAMETER *p;

  if(tree == misc && misc_buttons(handle, ob))
    return;
  if(ob == graph_ob){
    graph_edit(handle, tree, ob, table, size, m_x, m_y);
    return;
  }
  p = find_parameter(table, size, ob);
  if(!p)
    return;
  para_edit(handle, tree, ob, p->type, (data[10] == 1) ? p->d_page : p->s_page,
            p->position, data + p->offset);
  if(graph_ob == NO_GRAPH)
    return;
  g = (GRAPH *) (((USERBLK *) tree[graph_ob].ob_spec)->ub_parm);
  point = g->point;
  update = FALSE;
  for(i = 0; i < g->n_points; i++){
    if(point[i].x_ob == ob){
      t = data[p->offset];
      t = (point[i - 1].c_x + TIME2WIDTH(t)) - point[i].c_x;
      if(t != 0){
        update = TRUE;
        for(j = i; j < g->n_points && point[j].x_ob != FIXED; j++)
          point[j].c_x += t;
      }
      break;
    }
    if(point[i].y_ob == ob){
      t = data[p->offset];
      t = g->x_axis_y - ((p->type == UNUM) ? UNUM2HEIGHT(t) : SNUM2HEIGHT(t));
      if(t != point[i].c_y){
        update = TRUE;
        point[i].c_y = t;
        if(i + 1 < g->n_points && point[i + 1].y_ob == FOLLOW)
          point[i + 1].c_y = t;
      }
      break;
    }
  }
  if(update)
    wm_ob_update(handle, tree, graph_ob);
}

#define X_SENSE		8
#define Y_SENSE		8

static void graph_edit(int handle, OBJECT *tree, int ob,
                       PARAMETER table[], int size, int m_x, int m_y)
{
  GRAPH *g = (GRAPH *) (((USERBLK *) tree[ob].ob_spec)->ub_parm);
  CTRL_POINT *point = g->point;
  int i, j, x_offset, y_offset, umin_x, umax_x, link, n_move, t;
  GRECT update;
  VRECT bound;
  PARAMETER *p;

  objc_offset(tree, ob, &x_offset, &y_offset);
  bound.v_x1 = m_x - x_offset - X_SENSE;
  bound.v_x2 = m_x - x_offset + X_SENSE;
  bound.v_y1 = m_y - y_offset - Y_SENSE;
  bound.v_y2 = m_y - y_offset + Y_SENSE;
  for(i = 0;; i++){
    if(!(i < g->n_points))
      return;
    if(point[i].x_ob > FIXED || point[i].y_ob > FIXED)
      if(INSIDE(point[i].c_x, point[i].c_y, bound))
        break;
  }

  link = FALSE;
  if(i == 0){
    bound.v_x1 = point[i].c_x;
    bound.v_x2 = point[i].c_x;
    umin_x = point[i].c_x - X_SENSE;
    umax_x = MAX(point[i].c_x + X_SENSE, point[i + 1].c_x);
    n_move = 1;
  }
  else{
    bound.v_x1 = point[i - 1].c_x;
    bound.v_x2 = point[i - 1].c_x + T_WID;
    umin_x = bound.v_x1 - X_SENSE;
    umax_x = bound.v_x2 + X_SENSE;
    for(j = i + 1; j < g->n_points; j++)
      if(point[j].x_ob == FIXED){
        if(j == i + 1 && point[j].y_ob == FOLLOW){
          /*
           *	point[j].c_y follows point[i].c_y
           */
          link = TRUE;
          if(j + 1 < g->n_points)	/* point after follower exists */
            umax_x = point[j + 1].c_x;
          else
            umax_x = point[j].c_x;
        }
        else	/* point[j].c_y = point[j - 1].c_y are constant */
          umax_x = point[j - 1].c_x + (bound.v_x2 - point[i].c_x) + X_SENSE;
        break;
      }
    n_move = j - i;
  }
  if(point[i].y_ob == FIXED){
    bound.v_y1 = point[i].c_y;
    bound.v_y2 = point[i].c_y;
  }
  else{
    bound.v_y1 = g->y_min;
    bound.v_y2 = g->y_max;
  }
  update.g_x = x_offset + umin_x;
  update.g_w = umax_x - umin_x + 1;
  update.g_y = y_offset + g->y_min - Y_SENSE;
  update.g_h = g->y_max - g->y_min + 2 * Y_SENSE + 1;

  g->vsync_flag = TRUE;
  graph_drag(handle, tree, ob, x_offset, y_offset, &(point[i]), n_move, link,
             &bound, &update);
  g->vsync_flag = FALSE;

  t = SUCCESS;
  ob = point[i].x_ob;
  if(ob  > FIXED){
    p = find_parameter(table, size, ob);
    j = point[i].c_x - point[i - 1].c_x;
    j = WIDTH2TIME(j);
    if(j != data[p->offset]){
      data[p->offset] = j;
      t = tx_para_chg((data[10] == 1) ? p->d_page : p->s_page, p->position, j);
      unum_update(tree + ob, j);
      wm_ob_update(handle, tree, ob);
    }
  }
  ob = point[i].y_ob;
  if(ob  > FIXED){
    p = find_parameter(table, size, ob);
    j = g->x_axis_y - point[i].c_y;
    j = (p->type == UNUM) ? HEIGHT2UNUM(j) : HEIGHT2SNUM(j);
    if(j != data[p->offset]){
      data[p->offset] = j;
      if(t == SUCCESS)	/* don't generate two sysex failed alerts */
        tx_para_chg((data[10] == 1) ? p->d_page : p->s_page, p->position, j);
      if(p->type == UNUM)
        unum_update(tree + ob, j);
      else
        snum_update(tree + ob, j);
      wm_ob_update(handle, tree, ob);
    }
  }
}

static void graph_drag(int handle, OBJECT *tree, int ob,
                       int x_offset, int y_offset,
                       CTRL_POINT coord[], int n_move, int link,
                       VRECT *bound, GRECT *area)
{
  int i, m_x, m_y, new_x, new_y, m_state, k_state, redraw, delta;

  (void) intersect(&desk, area);
  wind_update(BEG_MCTRL);
  for(;;){
    graf_mkstate(&m_x, &m_y, &m_state, &k_state);
    if(m_state == 0)
      break;
    new_x = m_x - x_offset;
    if(new_x > bound->v_x2)
      new_x = bound->v_x2;
    else if(new_x < bound->v_x1)
      new_x = bound->v_x1;
    new_y = m_y - y_offset;
    if(new_y > bound->v_y2)
      new_y = bound->v_y2;
    else if(new_y < bound->v_y1)
      new_y = bound->v_y1;
    redraw = FALSE;
    delta = new_x - coord[0].c_x;
    if(delta != 0){
      for(i = 0; i < n_move; i++)
        coord[i].c_x += delta;
      redraw = TRUE;
    }
    if(new_y != coord[0].c_y){
      coord[0].c_y = new_y;
      if(link)
        coord[1].c_y = new_y;
      redraw = TRUE;
    }
    if(redraw)
      wm_ob_redraw(handle, tree, ob, area);
  }
  wind_update(END_MCTRL);
}

static int graph_draw(PARMBLK *pb)
{
  int x = pb->pb_x, y = pb->pb_y;
  GRAPH *g = (GRAPH *) pb->pb_parm;
  int i, *p, point[14];
  CTRL_POINT *q;

  point[0] = pb->pb_xc;
  point[1] = pb->pb_yc;
  point[2] = pb->pb_xc + pb->pb_wc - 1;
  point[3] = pb->pb_yc + pb->pb_hc - 1;
  vs_clip(graph_handle, 1, point);

  point[0] = x;
  point[1] = y;
  point[2] = x + pb->pb_w - 1;
  point[3] = y + pb->pb_h - 1;
  if(g->vsync_flag)
    Vsync();
  vr_recfl(graph_handle, point);

  vsl_type(graph_handle, 7);
  vsl_color(graph_handle, BLACK);
  point[0] = x + g->x_axis_x1;
  point[2] = x + g->x_axis_x2;
  point[1] = point[3] = y + g->x_axis_y;
  v_pline(graph_handle, 2, point);
  point[0] = point[2] = x + g->y_axis_x;
  point[1] = y + g->y_axis_y1;
  point[3] = y + g->y_axis_y2;
  v_pline(graph_handle, 2, point);
  point[0] = point[2] = x + g->y_axis2_x;
  v_pline(graph_handle, 2, point);

  vsl_type(graph_handle, 1);
  vsl_color(graph_handle, GREEN);
  for(i = g->n_points, p = point, q = g->point; i; i--, q++){
    *p++ = x + q->c_x;
    *p++ = y + q->c_y;
  }
  v_pline(graph_handle, g->n_points, point);
  
  for(i = g->n_points, p = point, q = g->point; i; i--, q++){
    if(q->x_ob > FIXED || q->y_ob > FIXED){
      *p++ = x + q->c_x;
      *p++ = y + q->c_y;
    }
  }
  v_pmarker(graph_handle, 4, point);
  
  return 0;
}

static PARAMETER *find_parameter(PARAMETER table[], int size, int ob)
{
  PARAMETER *p;

  for(p = table; p < table + size; p++)
    if(p->ob_number == ob)
      return p;
  return NULL;
}
  
static void para_setup(OBJECT *tree, PARAMETER table[], int size)
{
  PARAMETER *p;

  for(p = table; p < table + size; p++)
    para_fill(tree + p->ob_number, p->type, data + p->offset);
}

static int misc_buttons(int handle, int ob)
{
  switch(ob){
  case MS_COMPARE:
    if(!wm_do_button(handle, misc, ob, TRUE))
      return TRUE;
    if(misc[MS_COMPARE].ob_state & SELECTED){
      memcpy(compare, data, (size_t) PROGRAM_SIZE);
      memcpy(data, PROG_DATA(edit_bank, number), (size_t) PROGRAM_SIZE);
    }
    else
      memcpy(data, compare, (size_t) PROGRAM_SIZE);
    misc[MS_RESET].ob_state ^= DISABLED;
    misc[MS_SAVE].ob_state ^= DISABLED;
    break;
  case MS_RESET:
    if(!wm_do_button(handle, misc, ob, FALSE))
      return TRUE;
    memcpy(data, PROG_DATA(edit_bank, number), (size_t) PROGRAM_SIZE);
    break;
  case MS_SAVE:
    if(wm_do_button(handle, misc, ob, FALSE))
      save_prog(data);
    return TRUE;
  default:
    return FALSE;
  }
  tx_prog_dmp(data);
  prog_setup();
  wm_update(handle);
  return TRUE;
}

void wave_update(OBJECT *object, int value)
{
  object->ob_spec = (unsigned long) wave_image[value];
}
